/**

MM_4D_Writer -- A plugin that creates a 4d QuickTime movie
from a 4d set of TIFFs generated in MicroManager 1.4x.

By Jeff Hardin, Univ Of Wisconsin
		
Adapted from QT4DWriter, a plugin by Jeff Hardin
        
Some of the basic code was adapted from the MovieWriter plugin written by Wayne Rasband, NIH
        
The CompressionSettings class was modified from code by Charles Thomas, LOCI project,
Univ of Wisconsin-Madison.
        
Progress bar code was adapted from an online course hosted by Dave Marshall at
http://www.cs.cf.ac.uk/Dave/HCI/HCI_Handout_CALLER/node125.html
        
The basics of the image compression was adapted from Brad Lowe brad@nuspectra.com
based on various QT sample code

Kludgy workaround for bugs in the ImageCompressionDialog.requestSequenceSettings() call in
QTJ 6.1/Java 1.4.1 added
	
v. 0.1 - May 25, 2014
 */


import ij.plugin.frame.*;
import ij.*;
import ij.process.*;
import ij.io.*;
import ij.gui.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.util.Vector;
import java.lang.Number;

import javax.swing.*;
import javax.swing.event.*;

import quicktime.qd.*;
import quicktime.*;
import quicktime.std.*;
import quicktime.io.*;
import quicktime.std.image.*;
import quicktime.std.movies.*;
import quicktime.std.movies.media.*;
import quicktime.util.*;
import quicktime.app.view.*;
import quicktime.app.QTFactory;
import quicktime.std.qtcomponents.*;
 
public class MM_4D_Writer extends PlugInFrame implements StdQTConstants, Errors
    {
        public int slices,imageCount;
        public String firstImage;

        private int numFrames;
        private int kWidth;
        private int kHeight;
        private File soundFile;
        private String title;
        private ImageStack stack;
        public int myIndex;
        public boolean cancelOperation = false;
        public boolean workaround = false;

        //***** Compressor settings
        static int comp_type = kVideoCodecType;
        static int spatial_quality = codecNormalQuality;
        static float frame_rate = 8;
        static int key_frame_rate = 10;
        public int depth;
        
        //*** Crunch stuff
        QTFile movieDest;	// movie
        int codecQuality = codecNormalQuality;
        CodecName codecName = null;
        int maxFrames = -1;
        int curFrame = 0;
        QTFile images[];
        boolean flatten = false;
        static String defaultCodecName = "Animation";
        int kVidTimeScale = 600;	// 1 second = 600
        int defaultFrameDuration = 60;	// frames / 600 = 1/10 sec
        static boolean qtOpen = false;
        static String shortVersion = "ImageListToMovie dev 0.1 by brad@nuspectra.com";
        // List of image types to allow for importing.
        // more can be added.. keep lowercase.
        final static String okTypes[] = {".jpg", ".jpeg", ".gif", ".tif", ".tiff", ".bmp", ".png", ".pct"};
        int numPlanes = -1;
        double timeInterval = -1;
        String trimmedSpace = "";
        String movieInfo = null;
        ImagePlus img;

						
        public MM_4D_Writer()
        {
            super("Temp");
        }

        public void run(String arg)
        {
            //Check for Java version and QT version; need workaround for recent versions...
            String mySystemString = System.getProperty("java.version");
            double myQTJVersion = (double) QTBuild.getVersion() +  (double) QTBuild.getSubVersion()/10;
            //IJ.showMessage("Java VM version: " + mySystemString + "  QT Java version: " + QTBuild.getVersion() + "." + QTBuild.getSubVersion());
            //IJ.showMessage("Offset to version number: " + mySystemString.indexOf("1.4");
            //following updated to account for JVM 1.7
            if (((mySystemString.indexOf("1.4") >= 0) || (mySystemString.indexOf("1.5") >= 0) || (mySystemString.indexOf("1.6") >= 0) || (mySystemString.indexOf("1.7") >= 0)) && (myQTJVersion >= 6.1))
            {
                // Systems with QTJ 6.1 & Java VM 1.4 or higher require a workaround for now, so set boolean to true.
                workaround = true;
            }
            saveAsMovie();
        }


        public void parseData(String thePath)
        // This looks for micro-Manager metadata, and tries to read it.
        //  This just looks for focal planes
        {
		// Test for existence of file first!!
		String theName="metadata.txt";
		boolean infoFileExists = false;
		File myFile = new File(thePath,theName);		
		if (myFile.exists())
            {
                infoFileExists = true;
            }
		
		if (myFile.exists())
            {
                try
                {
				BufferedReader in = new BufferedReader(new FileReader(myFile));
				String str = "";
				String sliceString = ""; 
				String timeString = ""; 
				String spaceString = "";
				String channelString = "";
				int lineCount = 0;
				int startIndex = 0;
				boolean slicesNotFound = true; 
				boolean timeNotFound = true;
				while ((slicesNotFound) && ((str = in.readLine()) != null))
				{
                    lineCount = lineCount + 1;
                    if (str.lastIndexOf("Slices") > 0)
                        {
                            sliceString = str;
                            startIndex = sliceString.lastIndexOf(":");
                            sliceString = sliceString.substring(startIndex+1,sliceString.length()-1);
                            String trimmedSlices = sliceString.trim();
					    slices = Integer.parseInt (trimmedSlices);
                            //IJ.log("slices:" + slices);
                            slicesNotFound = false;
                            in.close();
                        }
                }//while
				BufferedReader in2 = new BufferedReader(new FileReader(myFile));
				lineCount = 0;
                startIndex = 0;
                str="";
				while ((timeNotFound) && ((str = in2.readLine()) != null))
				{
				lineCount = lineCount + 1;
				if (str.lastIndexOf("Interval_ms") > 0)
					{
					//IJ.log("Found interval string:" + str.lastIndexOf("Interval_ms"));
					timeString = str;
					//IJ.log("String: " +str);
					startIndex = timeString.lastIndexOf(":");
					timeString = timeString.substring(startIndex+1,timeString.length()-1);
					String trimmedTime = timeString.trim();
					//IJ.log("time: " + trimmedTime);
					//need to create double for timeInterval here!!!
					timeInterval = Double.valueOf(trimmedTime);
					}
				}//while

				in2.close();

				BufferedReader in3 = new BufferedReader(new FileReader(myFile));
				lineCount = 0; 
				startIndex = 0;
				str="";
				while ((timeNotFound) && ((str = in3.readLine()) != null))
				{
				lineCount = lineCount + 1;
				if (str.lastIndexOf("z-step_um") > 0)
					{
					//IJ.log("Found interval string:" + str.lastIndexOf("z-step_um"));
					spaceString = str;
					//IJ.log("String: " +str);
					startIndex = spaceString.lastIndexOf(":");
					spaceString = spaceString.substring(startIndex+1,spaceString.length()-1);
					trimmedSpace =spaceString.trim();
					//IJ.log("spacing: " + trimmedSpace);
					}
				}//while

				in3.close();

			}//try
			catch (IOException e)
			{
				IJ.log("There was an error parsing the XML file...");
			}
		}//if myFile.exists
    }//void

	public void saveAsMovie() 
	{
    //open folder with the images
    String dir = null;
    String foundName = null;
    String dstFile = "out.mov";
    String codec = defaultCodecName;
    int frameDuration = 0;
    Vector myList = new Vector();	// a list of image files to convert
    int quality = -1;
    boolean all = false;
	String infoFileRootName = null;
	slices=0; imageCount=0;
	DirectoryChooser dd = new DirectoryChooser("Open the dataset directory");
	if (dd==null) return;
	String dir1 = dd.getDirectory();
	parseData(dir1);		
    String[] list = new File(dir1).list();
	String[] list3= new String[list.length];
        if (list==null) {return;}
    // IJ.log("Parsing file data in the directory...");
	for (int i=0; i<list.length; i++) 
		{
		//IJ.log((i+1)+": "+list[i]);
		//IJ.showStatus((i+1) +"/"+list.length);
        //IJ.showProgress((i+1)/list.length);
		File f = new File(dir1+list[i]);
		if (!f.isDirectory()) 
			{
			if ((list[i].lastIndexOf("tif") > 0) || (list[i].lastIndexOf("jpg") > 0) || (list[i].lastIndexOf("gif") > 0) || (list[i].lastIndexOf("png") > 0) || (list[i].lastIndexOf("bmp") > 0)  || (list[i].lastIndexOf("raw") > 0)) 
				{
					imageCount = imageCount + 1; 
					list3[imageCount-1]= list[i];
				}
				if (imageCount == 1) {firstImage = list[i];
				ImagePlus img = new Opener().openImage(dir1,list[i]);
				}
			}//not a directory
		}//for loop

		IJ.log("\\Clear");
		//IJ.log("Images found: " + imageCount);
		//IJ.log("Slices: " + slices);
		//IJ.log("Time points: " + imageCount/slices);
		numPlanes = slices;

        if (list==null) {return;}
        // if (numPlanes > 0) {IJ.showMessage("Number of focal planes based on file info: " + (int)numPlanes);}
        // if (timeInterval < 0) { timeInterval = IJ.getNumber("Enter the number of seconds between time points: ",60);}
        //if (spaceInterval < 0) { spaceInterval = IJ.getNumber("Enter the slice spacing in microns between focal planes: ", 1);}
        if (movieInfo == null)
            {
            movieInfo = IJ.getString("Enter descriptive information about the movie: ","");
            }
		IJ.log("Movie information");
        IJ.log("-------------------");
        IJ.log("Number of focal planes: " + numPlanes);
        IJ.log("Seconds between time points: " + timeInterval/1000);
        IJ.log("Spacing between focal planes in microns: " + trimmedSpace);
        int totalTimePoints = 0;
		totalTimePoints = imageCount/slices;
		IJ.log("Total time points: " + totalTimePoints);
        IJ.log("Other info: " + movieInfo);
        Vector list2 = new Vector();
		int k = -1;
        myIndex = 0;
		for (int i=0; i<numPlanes; i++) 
			{
            // IJ.log("Focal plane " + (i+1) + ":");
            for (int j=0; j<totalTimePoints; j++)
            	{
                    k = k+1;
                    myIndex = j * numPlanes + i;
                    list2.addElement(new QTFile(new File(dir1, list3[myIndex])));
                    //IJ.log("Adding file " + list3[myIndex]);
                    //list2.add(myList[myIndex]);
                    //IJ.write("list2["+k+"]=" + list2[k]);
                    }
            }//for loop
		//stack = imp.getStack();
		//title = img.getTitle();
		try {
			QTSession.open();
			//initialize(myImg);
			//makeMovie();
			//canv.removeClient();
            // we have a list of image files to convert into a movie
            if (list2.size() > 0) 
            {
            	// IJ.showMessage("list2 has a non-zero size of " + list2.size());
            	QTFile imageFiles[] = new QTFile[list2.size()];
            	list2.copyInto(imageFiles);
            	//IJ.showMessage("imageFiles has a size of " + imageFiles.length);
            	MMCrunch converter = new MMCrunch(imageFiles);
				getCodecSettings(imageFiles[0]);
				//IJ.showMessage("cancelOperation = " + cancelOperation);
                if (cancelOperation == true) {return;}
            	//IJ.showMessage("Returned from image compression dialog!");
           		//codec = comp_type;
            	quality = spatial_quality;
				/*
            	IJ.log("");
		    	IJ.log("Compressor settings");
           	 	IJ.log("-------------------");
            	IJ.log("Quality: "+quality);
            	IJ.log("Codec type: "+comp_type);
            	IJ.log("Bit depth: " + depth);
            	IJ.log("Frame rate: " + frame_rate);
            	IJ.log("Key frame rate: "+key_frame_rate);
				**/
            	if (quality>=0) converter.setQuality(quality);
            	//converter.setCodec(codec);
            	converter.setCodec(comp_type);
            	frameDuration = (int) ((1/frame_rate)*600);
            	if (frameDuration > 0) converter.setFrameDuration(frameDuration);
            	title = "untitled";
            	SaveDialog sd = new SaveDialog("Save QuickTime Movie as...", title, ".mov");
				dir = sd.getDirectory();
				String name = sd.getFileName();
				if (name == null) throw new QTIOException(userCanceledErr, "");
				dstFile= dir+name;
				/*
            	IJ.log("");
            	IJ.log("Compressing image sequence...(this may take a while!)");
            	IJ.log("");
				**/
            	if (!all) converter.convert(new QTFile(dstFile));
            	writeInfoFile(dstFile+".txt");
            }

			QTSession.close();
			dispose();
                        
		} 
		catch (NoClassDefFoundError e) 
			{
			IJ.error("QuickTime for Java required");
			} 
		catch (Exception e) 
			{
			QTSession.close();
			}	
    }//saveAsMovie
        
    void writeInfoFile(String myFile)
        {
        
            //write the data to an info file - Line 1 = general info;
            //Line 2 = focal planes; Line 3 = space between focal planes
            //Line 4 = time interval
                
            FileOutputStream out; // declare a file output object
            PrintStream p; // declare a print stream object

            try
            {
                // Create a new file output stream
                // connected to "myfile.txt"
                out = new FileOutputStream(myFile);
                // Connect print stream to the output stream
                p = new PrintStream( out );
                p.println (movieInfo); p.println(numPlanes);
                p.println(trimmedSpace); p.println(timeInterval/1000);
                p.close();
            }
            catch (Exception e)
            {System.err.println ("Error writing to file");}
                
        }


    void printStackTrace(Exception e)
        {
		CharArrayWriter caw = new CharArrayWriter();
		PrintWriter pw = new PrintWriter(caw);
		e.printStackTrace(pw);
		IJ.write(caw.toString());
        }

    void getCodecSettings(QTFile test_file) throws Exception 
        {
		ImageCompressionDialog comp_dlog = null;
		SpatialSettings spatial_settings = null;
		TemporalSettings temporal_settings = null;
		DataRateSettings data_rate_settings = null;
		QDRect test_rect = null;
		int color_depth = 0; // let ICM choose depth
		CodecComponent comp = CodecComponent.anyCodec;
		int temporal_quality = codecNormalQuality;
		int data_rate = 1000;
		int frame_duration = 8;
		boolean cancelled = false;
		int test_flags = 0;
		int flags = 0;
        PixMap test_pix_map = null;
		
		//***** Get the standard QT compression dialog
		comp_dlog = new ImageCompressionDialog();
		//***** Set the default image                    
        //***** Set the default image
        //comp_dlog.setTestImagePixMap(test_pix_map, test_rect, test_flags);
        //IJ.showMessage("File: "+ test_file.getName());
                 
        GraphicsImporter myImporter = new GraphicsImporter(test_file);
        comp_dlog.setTestImagePict(myImporter.getAsPicture(), test_rect, test_flags);
                                                                        		   
		//***** Set up the default settings
		spatial_settings = new SpatialSettings(comp_type, comp, color_depth, spatial_quality);
		comp_dlog.setInfoSpatialSettings(spatial_settings);
		temporal_settings = new TemporalSettings(temporal_quality, frame_rate, key_frame_rate);
		comp_dlog.setInfoTemporalSettings(temporal_settings);
		data_rate_settings = new DataRateSettings(data_rate, frame_duration, codecMinQuality, codecMinQuality);
		comp_dlog.setInfoDataRateSettings(data_rate_settings);
		
		//***** Set the default such that BEST DEPTH is not an option
		flags &= ~scShowBestDepth;
		comp_dlog.setInfoPreferences(flags);
		cancelOperation = true;

		//***** Display the dialog for the user
		try
            {
			//requestSequenceSettings is broken with QTJ 6.1 & Java VM 1.4.1 -- needs a silly workaround
			//comp_dlog.requestSequenceSettings();
            if (workaround) {comp_dlog.requestImageSettings();} else {comp_dlog.requestSequenceSettings();}
            }
        catch (Error er)
            {
			throw new Error("User cancelled");
            }
			
		if (!cancelled) 
			{
			//***** Get the user settings
			spatial_settings = comp_dlog.getInfoSpatialSettings();
            depth = spatial_settings.getDepth();
			comp_type = spatial_settings.getCodecType();
			spatial_quality = spatial_settings.getSpatialQuality();	
            //following is broken!
            if (!workaround)
            	{	
				temporal_settings = comp_dlog.getInfoTemporalSettings();
				frame_rate = temporal_settings.getFrameRate();
                key_frame_rate = temporal_settings.getKeyFrameRate();
                cancelOperation = false;
            	}           
        //use dialog to circumvent
        //IJ.showMessage("Cancelled: " + cancelled + " workaround: " + workaround);
			else
				{
            	GenericDialog gd = new GenericDialog("Frame Rate Options", IJ.getInstance());
            	double xinitial = 8.0;
            	double yinitial = 0.0;
            	gd.addNumericField("Frame rate (fps):",xinitial, 1);
            	gd.addNumericField("Key frames:", yinitial,0);
				gd.showDialog();
				if (gd.wasCanceled()) {cancelOperation = true;} else{		
            	frame_rate = (float) gd.getNextNumber();
            	key_frame_rate = (int) gd.getNextNumber();
				cancelOperation = false;
				//IJ.showMessage("Cancel = false");
				}
		}//not cancelled
    }
}//void

        
public class MMCrunch
	{
	QTFile images[];
	boolean stopProgram;
	JFrame myFrame;
	JLabel myLabel;
	JProgressBar progress;

	public MMCrunch(QTFile images[]) throws QTException
		{
		init(images);
		}

	protected void init(QTFile images[]) throws QTException
		{
		//openQT();
		this.images = images;
		}

	public void setCodec(int myCType) throws Exception
		{
		CodecNameList list = new CodecNameList(1);
		for (int i=0;i<list.getCount();i++)
			{
			CodecName item = list.getNth(i);
			if (myCType == item.getCType())
				{
				codecName=item;
				return;
				}
			}
		}

	// spatial and temporal quality. 
	public void setQuality(int i)
		{
		codecQuality=i;
		}

	/**
	public void setCodec(int zIndex) throws Exception
		{
		CodecNameList list = new CodecNameList(1);
		codecName = list.getNth(zIndex);
		}
	*/
	public void setFrameDuration(int i)
		{
		defaultFrameDuration = i;
		}

	public int getMovieFrameCount() throws Exception
		{
		return images.length;
		}

	protected QTFile getNextImage() throws Exception
		{
		//IJ.showStatus("Compressing frame " + (curFrame+1) + " of " + images.length);
		//IJ.showProgress((curFrame+1)/images.length);
		return images[curFrame];
		}

	public void convert(File movieDest) throws Exception
		{
		convert(new QTFile(movieDest));
		}

	public void convert(QTFile movieDest) throws Exception        
        {
		//IJ.showMessage("Codec name: " + codecName.getName());
		if (codecName == null) setCodec(kVideoCodecType);
		stopProgram = false;

		Movie theMovie = Movie.createMovieFile (movieDest,
		kMoviePlayer, 
		createMovieFileDeleteCurFile | createMovieFileDontCreateResFile);

		//make window for display progress and cancel button
		progress = new JProgressBar();
		JButton button = new JButton();
    	//JLabel myLabel = new JLabel();
    	JPanel topPanel = new JPanel();
		myFrame = new JFrame("Compressing images...");
    	//myLabel = new Label("");
    	myFrame.setSize( 310, 130 );
		//setBackground( Color.gray );
		topPanel = new JPanel();
		topPanel.setPreferredSize( new Dimension( 310, 130 ) );
		myFrame.getContentPane().add( topPanel );

		// Create a label and progress bar
		myLabel = new JLabel( "" );
		myLabel.setPreferredSize( new Dimension( 280, 24 ) );
    	myLabel.setForeground(Color.black);
		topPanel.add( myLabel );

		progress = new JProgressBar();
		progress.setPreferredSize( new Dimension( 300, 20 ) );
		progress.setMinimum( 0 );
		progress.setMaximum( 90 );
		progress.setValue( 0 );
		progress.setBounds( 20, 35, 260, 20 );
    	progress.setForeground(Color.blue);
		topPanel.add( progress );

		button = new JButton( "Stop Compression" );
		topPanel.add( button );
		button.addActionListener(new ActionListener()
			{
        	public void actionPerformed(ActionEvent e) {
        	stopProgram = true;
        	myFrame.dispose();
        	}
        });
   	 	myFrame.show();

		try 
			{	
			addVideoTrack( theMovie );

			OpenMovieFile outStream = OpenMovieFile.asWrite (movieDest); 
			theMovie.addResource( outStream, movieInDataForkResID, movieDest.getName() );
			outStream.close();

			if (flatten)
				{
				// msTimer timer = new msTimer();
				theMovie.flatten(0, movieDest, kMoviePlayer, IOConstants.smSystemScript, createMovieFileDeleteCurFile | createMovieFileDontCreateResFile, movieInDataForkResID, movieDest.getName());
				// System.out.println (timer.reportString("flatten movie"));
				}
			myFrame.dispose();
			IJ.showStatus ("Finished movie. Duration:"+theMovie.getDuration() + " Length:"+movieDest.length());

			theMovie.disposeQTObject();
			theMovie = null;
			}
		catch (Exception qte) 
			{
			qte.printStackTrace(); 
			if (movieDest.exists()) movieDest.delete();
			throw qte;
			}

		}//void

	int getFrameDuration()
		{
		return defaultFrameDuration;	
		}

	private void addVideoTrack (Movie theMovie) throws QTException, Exception 
		{
		int numFrames = getMovieFrameCount();
		if (maxFrames > 0 && numFrames > maxFrames)	
		numFrames=maxFrames;
		progress.setMaximum(numFrames);//set the max for the progress bar
		QDRect rect=null;
		QDGraphics gw=null;
		Track vidTrack=null;
		VideoMedia vidMedia=null;
		QTHandle imageHandle=null;
		RawEncodedImage compressedImage=null;
		CSequence seq=null;
		ImageDescription desc=null;

		//IJ.write ("Compressing:" + numFrames + " frames");


		for (curFrame = 0; curFrame < numFrames; curFrame++) 
			{
			//IJ.log("f#:" + curFrame + "; about to get image...");
			QTFile imageFile = getNextImage();
			//IJ.log("f#:" + curFrame + "; got image...");
			// import the image into QuickTime
			GraphicsImporter myGraphicsImporter = new GraphicsImporter (imageFile);
			GraphicsImporterDrawer drawer = new GraphicsImporterDrawer (myGraphicsImporter);

			if (curFrame == 0)
				{
				/* Do initial setup, now that we can find out the width, height of first image */

				rect = myGraphicsImporter.getBoundsRect();
				gw = new QDGraphics (rect);
				vidTrack = theMovie.addTrack (rect.getWidth(), rect.getHeight(), 0);	// no volume
				vidMedia = new VideoMedia (vidTrack, kVidTimeScale); 
				vidMedia.beginEdits();

				int size = QTImage.getMaxCompressionSize (gw, rect, depth, codecQuality, codecName.getCType(), codecName.getCodecComponent());

				//IJ.log("Size: " + size);

				imageHandle = new QTHandle (size, true);
				imageHandle.lock();
				compressedImage = RawEncodedImage.fromQTHandle(imageHandle);
				seq = new CSequence (gw,rect, depth, codecName.getCType(), codecName.getCodecComponent(), codecQuality, codecQuality, MM_4D_Writer.key_frame_rate, null, 0);
				desc = seq.getDescription();
				}

			myLabel.setText("Compressing frame " +  (curFrame+1) + " of " + numFrames);
			progress.setValue( curFrame + 1 );

			if (stopProgram) {break;} 

			drawer.setGWorld (gw);
			drawer.redraw (null);

			CompressedFrameInfo info = seq.compressFrame (gw, rect, codecFlagUpdatePrevious, compressedImage);

			boolean isKeyFrame = info.getSimilarity() == 0;

			//if (curFrame % 10 == 0)
			//IJ.write("f#:" + curFrame + " kf=" + isKeyFrame + " sim=" + info.getSimilarity());
			//IJ.log("f#:" + curFrame + "; about to add sample...");
			vidMedia.addSample (imageHandle, 0, info.getDataSize(), getFrameDuration(), desc, 1, (isKeyFrame ? 0 : mediaSampleNotSync)); // no flags
			//IJ.log("f#:" + curFrame + "; added sample...");
			myGraphicsImporter.disposeQTObject();
			//IJ.log("f#:" + curFrame + "; disposed QTObject...");
		}//for

		vidMedia.endEdits();

		int kTrackStart	= 0;
		int kMediaTime = 0;
		int kMediaRate	= 1;
		vidTrack.insertMedia (kTrackStart, kMediaTime, vidMedia.getDuration(), kMediaRate);
	}//void

    }//class
}
