import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;
import ij.io.*;
import java.io.*;
import java.io.IOException;

/** This plugin opens a series of images (compressed or not)
 using the Metadata.txt file from a Micromanager dataset (MM v. 1.4). 
 
 Written 2012/09/25 by Jeff Hardin, Univ. of Wisconsin-Madison
 */

public class MM_Images_to_Hyperstack implements PlugIn {

public int slices,images;
public String firstImage;

	public void run(String arg) {
		slices=0; images=0;
		DirectoryChooser dd = new DirectoryChooser("Open the dataset directory");
		if (dd==null) return;
		String dir = dd.getDirectory();
		parseData(dir);
		countImages(dir);
		IJ.log("\\Clear");
		IJ.log("Images found: " + images);
		IJ.log("Slices: " + slices);
		IJ.log("Time points: " + images/slices);
		IJ.run("Image Sequence...", "open=[" + dir+firstImage+"] number=" + images + " starting=1 increment=1 scale=100 file=[] or=[] use");
		ImagePlus imp = IJ.getImage();
		IJ.run(imp, "Stack to Hyperstack...", "order=xyczt(default) channels=1 slices=" + slices +  " frames=" + images/slices+ " display=Grayscale");
	}

public void parseData(String thePath)
    // This looks for micro-Manager metadata, and tries to read it.
   //  This just looks for focal planes
    {
		// Test for existence of file first!!
		String theName="metadata.txt";
		boolean infoFileExists = false;
		File myFile = new File(thePath,theName);		
		if (myFile.exists())
		{
			infoFileExists = true;
		}
		
		if (myFile.exists())
		{
			try
			{
				BufferedReader in = new BufferedReader(new FileReader(myFile));
				String str = "";
				String sliceString = "";
				String channelString = "";
				int lineCount = 0;
				int startIndex = 0;
				boolean slicesNotFound = true;
				while ((slicesNotFound) && ((str = in.readLine()) != null))
				{
					lineCount = lineCount + 1;
			if (str.lastIndexOf("Slices") > 0)
					{
						sliceString = str;
						startIndex = sliceString.lastIndexOf(":");
						sliceString = sliceString.substring(startIndex+1,sliceString.length()-1);
				String trimmedSlices = sliceString.trim();
				slices = Integer.parseInt (trimmedSlices);
				slicesNotFound = false;
					}
				}
				in.close();
			}
			catch (IOException e)
			{
				IJ.log("There was an error parsing the XML file...");
			}
		}
    }

	public void countImages(String dir1)
	{
               // IJ.log("Directory: " + dir1);
                String[] list = new File(dir1).list();
		if (list==null) return;                
              // IJ.log("Parsing file data in the directory...");
		for (int i=0; i<list.length; i++) {
			//IJ.log((i+1)+": "+list[i]);
			//IJ.showStatus((i+1) +"/"+list.length);
                        //IJ.showProgress((i+1)/list.length);
			File f = new File(dir1+list[i]);
			if (!f.isDirectory()) {
				if ((list[i].lastIndexOf("tif") > 0) || (list[i].lastIndexOf("jpg") > 0) || (list[i].lastIndexOf("gif") > 0) || (list[i].lastIndexOf("png") > 0) || (list[i].lastIndexOf("bmp") > 0)  || (list[i].lastIndexOf("raw") > 0)) images = images + 1; 
				if (images == 1) {firstImage = list[i];}
			}//not a directory
		}//for loop
		//should check for integer value for time points!
	}


}
